/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author(s): Damien Fournier
 *            Christophe Demarey
 */

package org.ow2.frascati.factory;

import java.io.File;
import java.net.URL;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.ow2.frascati.factory.core.dispatcher.Manager;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;
import org.ow2.frascati.factory.generate.runtime.RuntimeGenerate;

/**
 * The GenerateFactory generates stuff for a SCA composite description.
 * 
 * @author <a href="mailto:Damien.Fournier@inria.fr">Damien Fournier</a>
 * @author <a href="mailto:Lionel.Seinturier@lifl.fr">Lionel Seinturier</a>
 * @author Valerio Schiavoni <valerio.schiavoni@gmail.com>
 * @author Christophe Demarey
 */
public class GenerateFactory extends AbstractGenerateFactory {

  /**
   * Run the Factory with given arguments (composite path and the classpath).
   * This method is simply called by the main method of the FactoryCommandLine
   * tool.
   * 
   * @param factory
   *          - The factory instance to use (AssemblyFactory, RuntimeFactory,
   *          Generate Factory).
   * @param compositeName
   *          The name of the SCA composite to execute.
   * @param serviceName
   *          The name of the service to execute.
   * @param methodName
   *          The name of the method to execute.
   * @param classpath
   *          - List of URLs used to define the classpath.
   */

  public static void run(AbstractFactory factory, String compositeName,
      String serviceName, String methodName, URL[] classpath) {
    // This factory only generates code and can't run a SCA composite
    System.err.println("The method run(factory, compositeName, serviceName, "
        + "methodName, classpath) is NOT IMPLEMENTED in this factory!");
  }

  /**
   * Run the Factory with given arguments (composite path and the classpath).
   * This method is simply called by the main method of the FactoryCommandLine
   * tool.
   * 
   * @param compositeName
   *          The name of the SCA composite to execute.
   * @param libs
   *          - List of URLs used to define the classpath.
   * @param srcs
   *          - List of source files root folders.
   * @param output
   *          - The base output directory.
   */

  public static void run(String compositeName, URL[] libs, String[] srcs,
      String output) {

    GenerateFactory factory = new GenerateFactory(srcs);
    factory.setOutput(output);

    new GenerateFactory().getComposite(compositeName, libs);

  }

  private String[] srcs = { "/src/main/java", "/src/main/resources" };

  // Methods

  // Constructors
  public GenerateFactory() {
    super();
  }

  /**
   * Create a new Factory capable for compiling SCA source code
   * 
   * @param src
   *          - Paths for source code
   * 
   */
  public GenerateFactory(String[] srcs) {
    super();
    this.srcs = srcs;
  }

  /**
   * Generate and compile java source code according to an SCA composite file
   * 
   * @param composite
   *          - The composite file located into the current class path, on the
   *          local file system, or at an URL
   * @param libs
   *          - Libraries required to load this composite
   * 
   * @return null
   */

  public Component getComposite(String composite, URL[] libs) {
    log.info("Generating the SCA composite : " + composite);

    // Initialize Generate factory.
    try {

      // Get reference on the "juliac" interface given by the
      // Assembly factory runtime component
      RuntimeGenerate juliac = (RuntimeGenerate) getFactory().getFcInterface(
          "juliac");

      // Get reference on the "runtime" interface given by the
      // Assembly factory runtime component
      RuntimeFactory runtime = (RuntimeFactory) getFactory().getFcInterface(
          "runtime");

      // Get reference on the "load" interface given by the
      // Assembly factory runtime component
      Manager load = ((Manager) getFactory().getFcInterface("load"));

      // Init Juliac and set output for generated code
      // juliac.init();
      // juliac.setOutputDir(output.getCanonicalFile());

      // Add path to java source code
      // Use only with generate factory
      juliac.addJavaSource(srcs);

      URL[] srcUrl = new URL[srcs.length];
      for (int i = 0; i < srcs.length; i++) {
        srcUrl[i] = new File(srcs[i]).toURL();
      }

      // Load resources from srcs and into factory ClassLoader
      runtime.loadLibraries(srcUrl);

      // ClassLoader parent = getClassLoader();
      // URLClassLoader urlcl = new URLClassLoader(srcUrl, parent);
      // setClassLoader(urlcl);

      // Load additional libraries
      if (libs != null)
        load.loadJars(libs);

      // compile source code
      juliac.compile();

      // Parse SCA composite.
      load.loadComposite(composite);

      // compile generated source code
      juliac.compile();

      log.info("SCA composite: " + composite + " generated successfully");

    } catch (Exception e) {
      log.severe("Problem during factory initialization : " + e.getMessage());
      e.printStackTrace();
    }

    return null;
  }

  /**
   * Get the factory main class.
   */
  @Override
  protected String getFactoryMainClass() {
    return "org.ow2.frascati.factory.FrascatiGenerate";
  }

  /* (non-Javadoc)
   * @see org.ow2.frascati.factory.AbstractGenerateFactory#setOutput(java.lang.String)
   */
  public void setOutput(String output) {
    this.output = new File(output);

    // Get reference on the "juliac" interface given by the
    // Assembly factory runtime component

    try {
      RuntimeGenerate juliac = (RuntimeGenerate) getFactory().getFcInterface(
          "juliac");
      juliac.setOutputDir(this.output);

    } catch (NoSuchInterfaceException e) {
      throw new Error("Cannot get set Juliac output directory ", e);
    }

  }

}
