/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 */

package org.ow2.frascati.factory.core.instance.runtime;

import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.logging.Logger;

/**
 * Manage ClassLoader used by OW2 FraSCAti Assembly Factory and load composite
 * resources
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 */

public class Loader implements FactoryLoaderItf {

  /**
   * The used logger.
   */

  protected static Logger log = Logger.getLogger(Loader.class
      .getCanonicalName());

  /** File extension for SCA composites */
  public static final String SCA_FILENAME_EXT = ".composite";

  /** File extension for libraries */
  public static final String LIBRARY_EXT = ".jar";

  // initial ClassLoader for OW2 FraSCAti Assembly Factory
  protected static ClassLoader initClassloader;

  // ClassLoader currently used by OW2 FraSCAti Assembly Factory
  protected ClassLoader currentClassloader;
  
  /**
   * Constructor for Loader Class 
   */
  
  
  public Loader() {
    
    initClassloader = getClass().getClassLoader();
    currentClassloader = new AFClassLoader(new URL[0], initClassloader);
    
    // May be needed when using binding factory
    Thread.currentThread().setContextClassLoader(this.currentClassloader);

  }

  /* (non-Javadoc)
   * @see org.ow2.frascati.factory.core.instance.runtime.FactoryLoaderItf#getClassLoader()
   */
  
  public ClassLoader getClassLoader() {
    return this.currentClassloader;
  }

  /**
   * @see org.ow2.frascati.factory.core.instance.runtime.FactoryLoaderItf#setClassLoader(java.lang.ClassLoader)
   */

  public void setClassLoader(ClassLoader classloader) {

    this.currentClassloader = classloader;

    // May be needed when using binding factory
    // TODO: manage classloader between OW2 FraSCAti Assembly Factory and
    // Binding Factor
    
    Thread.currentThread().setContextClassLoader(this.currentClassloader);

  }

  /**
   * @see org.ow2.frascati.factory.core.instance.runtime.FactoryLoaderItf#getFactoryClassLoader()
   */

  public ClassLoader getFactoryClassLoader() {
    return initClassloader;
  }

  /**
   * @see org.ow2.frascati.factory.core.instance.runtime.FactoryLoaderItf#getCompositeResource(java.lang.String)
   */

  public URI getCompositeResource(String compositeFile) {
    try {
      // Composite URI
      URI compositeUri = null;

      // add ".composite" at the end of the filename if necessary
      if (!compositeFile.endsWith(SCA_FILENAME_EXT))
        compositeFile = compositeFile + SCA_FILENAME_EXT;

      // Try to load composite resource from the current class loader
      URL compositeUrl = currentClassloader.getResource(compositeFile);

      if (compositeUrl != null) {
        return compositeUrl.toURI();
      }

      // Composite file not loaded as resource of the current class loader
      // Try to load from the local file system or an URL.

      else {
        compositeUri = URI.create(compositeFile);
      }

      return compositeUri;

    } catch (URISyntaxException e) {
      throw new Error(
          "Could not retrieve valid URI for this composite resource");
    }

  }

  /* (non-Javadoc)
   * @see org.ow2.frascati.factory.core.instance.runtime.FactoryLoaderItf#loadLibraries(java.net.URL[])
   */
  
  public void loadLibraries(URL[] urls) {

    ClassLoader currentUrl = getClassLoader();

    if (currentUrl instanceof AFClassLoader) {

      // Current ClassLoader is an AFClassLoader
      AFClassLoader currentUrlCl = (AFClassLoader) currentUrl;

      // Add urls to current ClassLoader
      for (URL url : urls) {
        currentUrlCl.addUrl(url);
      }

    }

    else {
      // Create new AFClassLoader
      // Keep reference on parent
      ClassLoader libClassLoader = new AFClassLoader(urls, currentUrl);

      // Replace current ClassLoader by new one
      setClassLoader(libClassLoader);

    }

  }

  /**
   * Debug output for given ClassLoader - print loaded URLs
   * 
   * @param cl
   *          - ClassLoader used
   */
  public void debug(ClassLoader cl) {
    System.err.println("----- Debugging ClassLoader -----");
    if (cl == null)
      cl = currentClassloader;
    System.err.println("List of URL loaded for ClassLoader : " + cl);
    System.err.println("--------------------------------------");
    while (cl instanceof URLClassLoader) {
      URLClassLoader urlcl = (URLClassLoader) cl;
      for (URL url : urlcl.getURLs()) {
        System.err.println("> " + url);
      }

      cl = urlcl.getParent();

    }

  }

}
