/***
 * OW2 FraSCAti Assembly Factory
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org 
 *
 * Author: Damien Fournier
 */

package org.ow2.frascati.factory.core.parser.util;

import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.stp.sca.Component;
import org.eclipse.stp.sca.ComponentReference;
import org.eclipse.stp.sca.ComponentService;
import org.eclipse.stp.sca.Composite;
import org.eclipse.stp.sca.Implementation;
import org.eclipse.stp.sca.Interface;
import org.eclipse.stp.sca.JavaImplementation;
import org.eclipse.stp.sca.JavaInterface;
import org.eclipse.stp.sca.ScaFactory;
import org.eclipse.stp.sca.ScaPackage;
import org.eclipse.stp.sca.impl.ScaFactoryImpl;
import org.osoa.sca.annotations.Reference;
import org.osoa.sca.annotations.Service;
import org.ow2.frascati.factory.core.instance.runtime.RuntimeFactory;
import org.ow2.frascati.tinfi.metadata.MetaData;

/**
 * Process java classes to add missing services and references to the model
 * 
 * @author <a href="mailto:damien.fournier@inria.fr">Damien Fournier</a>
 * @version $Revision$
 * @param <T>
 */

public class JavaInterfaceResolver implements Resolver {

  /**
   * SCA Model factory
   */

  ScaFactory scaFactory = new ScaFactoryImpl();

  /**
   * Reference on runtime component
   */

  protected RuntimeFactory runtime;

  /**
   * Logger for this class
   */

  java.util.logging.Logger log = java.util.logging.Logger
      .getLogger(JavaInterfaceResolver.class.getCanonicalName());

  public JavaInterfaceResolver(RuntimeFactory runtime) {
    this.runtime = runtime;
  }

  /**
   * Resolve Java components implementation
   * 
   * @param composite
   * @return resulting composite
   */

  @SuppressWarnings("unchecked")
  public Composite resolve(Composite composite) {

    // Retrieve components
    Iterator<Component> componentIterator = composite.getComponent().iterator();

    while (componentIterator.hasNext()) {

      Component c = componentIterator.next();
      Implementation imp = c.getImplementation();
      if (imp.eClass().getClassifierID() == ScaPackage.JAVA_IMPLEMENTATION) {
        JavaImplementation javaImpl = (JavaImplementation) imp;
        try {

          Class<?> cl = runtime.getClassLoader()
              .loadClass(javaImpl.getClass_());
          MetaData<?> ccmd = new MetaData(cl);

          resolveServices(c, ccmd);
          resolveReferences(c, ccmd);

        } catch (ClassNotFoundException e) {
          log.fine("Could not get Meta Data from : " + javaImpl.getClass_());
        }

      }

    }

    return composite;

  }

  /**
   * Resolve service interfaces defined in java
   * 
   * @param component
   *          - current component
   * @param cMetaData
   *          - ContentClassMetaData API instance
   */

  public void resolveServices(Component component, MetaData<?> cMetaData) {

    // Retrieve service annotation
    Service s = cMetaData.getAnnotation(Service.class);
    Class<?>[] interfaces = null;

    // Test if class is annotated with org.osoa.sca.annotations.Service
    if (s != null) {

      // Array to store interfaces defined in java file
      if (s.interfaces().length < 1) {
        interfaces = new Class<?>[1];
        interfaces[0] = s.value();
      } else
        interfaces = s.interfaces();
    } else {
      // Class not annotated with org.osoa.sca.annotations.Service use
      // implemented interfaces instead
      // TODO test if implemented interfaces are annotated with @Remote
      JavaImplementation javaImpl = (JavaImplementation) component
          .getImplementation();
      try {
        Class<?> implementation = Class.forName(javaImpl.getClass_(), false,
            runtime.getClassLoader());
        interfaces = implementation.getInterfaces();
      } catch (ClassNotFoundException e) {
        log.severe("Class " + javaImpl.getClass_() + " may not be loaded");
        e.printStackTrace();
      }
    }

    // create current services map
    Map<String, String> serviceMap = new HashMap<String, String>();
    // create missing services map
    Map<String, String> missingServiceMap = new HashMap<String, String>();

    // register services already defined in the model for this component
    EList<ComponentService> serviceList = component.getService();
    for (ComponentService componentService : serviceList) {
      Interface itf = componentService.getInterface();
      // Test if service interface is defined
      if (itf != null) {
        // Test if service interface is a java interface
        if (componentService.getInterface().eClass().getClassifierID() == ScaPackage.JAVA_INTERFACE) {
          // register
          serviceMap.put(componentService.getName(),
              ((JavaInterface) componentService.getInterface()).getInterface());
        }
      }
      // Service is defined without interfaces, need to add if possible
      else {
        JavaInterface i = scaFactory.createJavaInterface();
        i.setInterface(interfaces[0].getCanonicalName());
        componentService.setInterface(i);
        // register since just added
        serviceMap.put(componentService.getName(), interfaces[0]
            .getCanonicalName());
      }
    }

    // register missing services
    for (Class<?> jInterface : interfaces) {
      // Test if interface type as already been registered
      if (!serviceMap.containsValue(jInterface.getName())) {
        String sName = jInterface.getSimpleName();
        missingServiceMap.put(sName, jInterface.getName());
      }
    }

    Set<String> serviceSet = missingServiceMap.keySet();
    for (String key : serviceSet) {
      // create a new component service to add
      ComponentService c = scaFactory.createComponentService();
      JavaInterface i = scaFactory.createJavaInterface();
      i.setInterface(missingServiceMap.get(key));
      c.setName(key);
      c.setInterface(i);
      // add component service to the model
      component.getService().add(c);
    }

  }

  /**
   * Resolve reference interfaces defined in java
   * 
   * @param component
   *          - current component
   * @param cMetaData
   *          - ContentClassMetaData API instance
   */

  public void resolveReferences(Component component, MetaData<?> cMetaData) {

    // Retrieve Reference annotation
    AccessibleObject o[] = cMetaData
        .getAllAnnotatedMethodsAndFields(Reference.class);

    // Test if fields or methods are annotated with
    // org.osoa.sca.annotations.reference
    if (o != null) {

      // create current references map
      Map<String, String> referenceMap = new HashMap<String, String>();

      // create and register references map from java class
      Map<String, String> referenceDecMap = new HashMap<String, String>();

      // create and register references annotated object from java class
      // TODO similar to referenceDecMap need to evaluate if it is better to
      // create a second map or use getter for retrieving reference
      // implementation class name

      // TODO Not Currently used
      // Map<String, AccessibleObject> referenceObjectMap = new HashMap<String,
      // AccessibleObject>();

      // register reference annotated fields and methods in map
      for (AccessibleObject object : o) {
        referenceDecMap.put(getReferenceName(object),
            getReferenceClassName(object));
        // TODO Not Currently used
        // referenceObjectMap.put(getReferenceName(object), object);
      }

      // register references already defined in the model for this component
      EList<ComponentReference> referenceList = component.getReference();
      for (ComponentReference componentReference : referenceList) {
        Interface itf = componentReference.getInterface();
        // Test if reference has a defined interface
        if (itf != null) {
          if (itf.eClass().getClassifierID() == ScaPackage.JAVA_INTERFACE) {
            // register
            referenceMap.put(componentReference.getName(),
                ((JavaInterface) componentReference.getInterface())
                    .getInterface());
          }
        }
        // references are defined without interfaces, need to add if possible
        else {
          // get interface from the reference defined in the implementation
          String referenceName = componentReference.getName();
          String interfaceClass = referenceDecMap.get(referenceName);
          if (interfaceClass != null) {
            JavaInterface newJavaItf = scaFactory.createJavaInterface();
            newJavaItf.setInterface(interfaceClass);
            componentReference.setInterface(newJavaItf);
            // put to list of registered references
            referenceMap.put(referenceName, interfaceClass);
          }
        }

        // resolveCallback(componentReference, cMetaData);

      }

      // register missing references
      Set<String> references = referenceDecMap.keySet();
      for (String key : references) {
        // Test if the reference (by name) defined in java class has been
        // defined in Model
        if (!referenceMap.containsKey(key)) {
          // This reference must be added to current model
          // create a new component service to add
          ComponentReference c = scaFactory.createComponentReference();
          JavaInterface i = scaFactory.createJavaInterface();
          i.setInterface(referenceDecMap.get(key));
          c.setName(key);
          c.setInterface(i);
          // add component service to the model
          component.getReference().add(c);

        }

      }

    }

  }

  /**
   * Get Class of a reference from annotated field / method
   * 
   * @param object
   *          - Reference annotated Field or Method
   * @return reference Class name
   */

  private String getReferenceClassName(AccessibleObject object) {

    if (object instanceof Field) {
      Field field = (Field) object;
      return field.getType().getCanonicalName();
    }

    if (object instanceof Method) {
      Method method = (Method) object;
      Class<?>[] classe = method.getParameterTypes();
      if (classe.length > 0)
        return classe[0].getCanonicalName();
    }

    return null;

  }

  /**
   * Get Name of a reference from annotated field / method
   * 
   * @param object
   *          - Reference annotated Field or Method
   * @return reference name
   */

  private String getReferenceName(AccessibleObject object) {

    if (object instanceof Field) {
      Field field = (Field) object;
      String name = field.getAnnotation(Reference.class).name();
      if (name.equals(""))
        name = field.getName();
      return name;
    }

    if (object instanceof Method) {
      Method method = (Method) object;
      String name = method.getAnnotation(Reference.class).name();
      if (name.equals("")) {
        name = method.getName().substring(3);
        name = name.substring(0, 1).toLowerCase()
            + name.substring(1, name.length());
      }
      return name;
    }

    return null;

  }

}
