/***
 * OW2 FraSCAti Juliac
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.Interceptor;
import org.objectweb.fractal.julia.factory.ChainedInstantiationException;
import org.objectweb.fractal.julia.loader.Initializable;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.ClassGenerator;
import org.objectweb.fractal.juliac.JuliacLoader;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.desc.ControllerDesc;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.desc.NoSuchControllerDescriptorException;
import org.objectweb.fractal.juliac.proxy.InterfaceImplementationClassGenerator;
import org.objectweb.fractal.juliac.runtime.JuliacBootstrapComponentImpl;
import org.objectweb.fractal.juliac.visit.MethodSourceCodeVisitor;

/**
 * The Juliac bootstrap component.
 * 
 * This bootstrap component provides a type-factory, a generic-factory and a
 * classloader interface.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 0.4.3
 */
public class TinfiBootstrapComponentImpl extends JuliacBootstrapComponentImpl {
        
    // -----------------------------------------------------------------
    // Implementation of the GenericFactory interface
    // -----------------------------------------------------------------
    
    /**
     * Create a new component.
     * 
     * This method tries either to load a previously generated component factory
     * (a so-called component initializer class) or to use the Juliac generic
     * component factory.
     * 
     * @param type            the component type
     * @param controllerDesc  the description of the control membrane associated to the component
     * @param contentDesc     the description of the content of the component
     */
    @Override
    public Component newFcInstance(
        Type type, Object controllerDesc, Object contentDesc )
    throws InstantiationException {
        
        try {
            Component c = super.newFcInstance(type,controllerDesc,contentDesc);
            return c;
        }
        catch( InstantiationException ie ) {
            if( ie instanceof ChainedInstantiationException ) {
                ChainedInstantiationException cie =
                    (ChainedInstantiationException) ie;
                Throwable t = cie.getException();
                if( ! (t instanceof ClassNotFoundException) ) {
                    /*
                     * Propagate unless this is a ClassNotFoundException.
                     */
                    throw ie;
                }
            }            
        }
        
        /*
         * A ClassNotFoundException has been thrown. This occurs when the
         * component factory (so-called component initializer class) can not be
         * loaded. In this case, use a generic component factory method. 
         */
        
        // Castability has been checked by super.newFcInstance
        ComponentType ct = (ComponentType) type;
        String ctrlDesc = (String) controllerDesc;

        if( contentDesc==null || contentDesc instanceof String ) {
            Component c = genericNewFcInstance(ct,ctrlDesc,contentDesc);
            return c;
        }
        else {
            Component c =
                genericNewFcInstance(ct,ctrlDesc,contentDesc,contentDesc);
            return c;
        }
    }
    
        
    // -----------------------------------------------------------------
    // Generic factory implementation
    // -----------------------------------------------------------------
    
    /**
     * Juliac generic factory component method.
     * 
     * @param ct           the component type
     * @param ctrlDesc     the description of the control membrane associated to the component
     * @param contentDesc  the description of the content of the component
     */
    protected Component genericNewFcInstance(
        ComponentType ct, String ctrlDesc, Object contentDesc )
    throws InstantiationException {
        
        Object content = newFcContent(contentDesc);
        Component c = genericNewFcInstance(ct,ctrlDesc,contentDesc,content);
        return c;
    }
    
    protected Object newFcContent( Object contentDesc )
    throws InstantiationException {
        
        if( contentDesc == null ) {
            return null;
        }
        else {
            if( !(contentDesc instanceof String) ) {
                String msg = "contentDesc should be a String";
                throw new InstantiationException(msg);
            }
            String name = (String) contentDesc;
            Class<?> cl = loadClassIE(name);
            Object content = instanciateClass(cl);
            return content;
        }
    }

    /**
     * The code for this method is inspired from that of 
     * {@link org.objectweb.fractal.juliac.opt.oo.InitializerOOCtrlClassGenerator#generateNewFcInstanceContentMethod(MethodSourceCodeVisitor)}
     * method in module org.objectweb.fractal.juliac:juliac-oo
     */
    protected Component genericNewFcInstance(
        ComponentType ct, String ctrlDesc, Object contentDesc, Object content )
    throws InstantiationException {
        
        ClassLoader classloader = getClassLoader();
        JuliacLoader loader = new JuliacLoader(classloader);
        Map<?,?> context = new HashMap<Object,Object>();
        
        /*
         * Get initial information about the membrane type, the list of
         * controller descriptors and the membrane descriptor.
         */        
        InterfaceType[] membraneits = null;
        List<ControllerDesc> ctrlDescs = null;
        try {
            membraneits = loader.getMembraneType(ctrlDesc,context);
            ctrlDescs = loader.getCtrlsImplLayers(ctrlDesc,context);        
        }
        catch (NoSuchControllerDescriptorException nscde) {
            throw new ChainedInstantiationException(nscde,null,"");
        }
        MembraneDesc membraneDesc =
            new MembraneDesc(null,ctrlDesc,membraneits,ctrlDescs);
        membraneDesc.setInterceptorClassGenerator(icg);  // TODO fix me
        
        // Check the content
        contentChecks(ct,content);
        
        // Prepare an InitializationContext
        InitializationContext ic = new InitializationContext();
        ic.controllers = new ArrayList<Object>();
        ic.interfaces = new HashMap<String,Object>();
        ic.internalInterfaces = new HashMap<String,Object>();
        
        // Instantiate the content
        initializationContextForContent(ic,content,ctrlDesc,contentDesc);
        
        // Instantiate controllers
        Component compctrl = null;
        for (ControllerDesc cd : ctrlDescs) {
            String ctrlImplName = cd.getImplName();
            if( ctrlImplName.charAt(0) == '(' ) {
                /*
                 * Skip CloneableAttributeController implementations.
                 * The controller implementation class name is in this case
                 * something like:
                 * 
                 * ((attribute-controller org.objectweb.fractal.julia.conform.components.CAttributes))
                 */
                continue;
            }
            Class<?> cl = loadClassIE(ctrlImplName);
            Object ctrl = instanciateClass(cl);
            ic.controllers.add(ctrl);
            if( Initializable.class.isAssignableFrom(cl) ) {
                Tree tree = cd.getTree();
                if( tree != null ) {
                    try {
                        ((Initializable)ctrl).initialize(tree);
                    }
                    catch (Exception e) {
                        throw new ChainedInstantiationException(e,null,"");
                    }
                }
            }
            if( ctrl instanceof Component ) {
                compctrl = (Component) ctrl;
            }
        }

        InterfaceType compctrlit =
            new BasicInterfaceType(
                "component",Component.class.getName(),false,false,false);

        // Interceptor (no interceptor for the component controller)
        Object intercept = null;
        intercept = compctrl;

        // Component controller
        Class<?> compFcItfCl = loadFcItfClass(compctrlit);
        Interface proxy =
            instanciateFcItfClass(
                compFcItfCl,compctrl,"component",compctrlit,false,intercept);            
        Component proxyForCompCtrl = (Component) proxy;
        ic.interfaces.put("component",proxy);
        
        boolean attrimplgenerated = false;

        // Controllers other than the component controller
        for (int i=1; i < membraneits.length; i++) {
            
            String itname = membraneits[i].getFcItfName();

            // Control external interface
            if( itname.equals("/cloneable-attribute-controller") ) {
                attrimplgenerated = true;
                // No such interface type in scaPrimitive and scaComposite membranes
                String msg =
                    "Unexpected case: no such interface type /cloneable-attribute-controller in scaPrimitive and scaComposite membranes";
                throw new InstantiationException(msg);
            }
            else {
                String signature = membraneits[i].getFcItfSignature();
                Class<?> cl = loadClassIE(signature);
                Object ctrl = null;
                for( int j=0 ; j < ic.controllers.size() ; j++ ) {
                    ctrl = ic.controllers.get(j);
                    if( cl.isAssignableFrom(ctrl.getClass()) ) {
                        break;
                    }
                    ctrl = null;
                }
                if( ctrl == null ) {
                    if( cl.isAssignableFrom(content.getClass()) ) {
                        ctrl = content;
                    }
                    else {
                        String msg =
                            "No such controller implementing "+signature;
                        throw new InstantiationException(msg);
                    }
                }
                
                // Interceptors (no interceptor for controllers)
                intercept = ctrl;
                Class<?> ctrlFcItfCl = loadFcItfClass(membraneits[i]);
                proxy =
                    instanciateFcItfClass(
                        ctrlFcItfCl, proxyForCompCtrl, itname, membraneits[i],
                        false, intercept );
            }
            ic.interfaces.put(itname,proxy);

            // Control internal interface
            if( itname.equals("factory") ) {
                InterfaceType intit =
                    Utils.newSymetricInterfaceType(membraneits[i]);
                Class<?> ctrlFcItfCl = loadFcItfClass(intit);
                proxy =
                    instanciateFcItfClass(
                        ctrlFcItfCl, proxyForCompCtrl, itname, intit, true,
                        intercept );
                ic.internalInterfaces.put(itname,proxy);
            }
        }

        // Component type
        InterfaceType[] its = ct.getFcInterfaceTypes();
        InterfaceType[] fullits = new InterfaceType[ membraneits.length + its.length ];
        System.arraycopy(membraneits, 0, fullits, 0, membraneits.length);
        System.arraycopy(its, 0, fullits, membraneits.length, its.length);
        ic.type = new BasicComponentType(fullits);
        
        // Business interfaces
        for (int i = 0; i < its.length; i++) {
            
            String itname = its[i].getFcItfName();
            String itsignature = its[i].getFcItfSignature();
            
            if( its[i].isFcCollectionItf() ) {
                // Julia naming convention for collection interfaces
                itname = "/collection/" + itname;
            }

            // Interceptors
            intercept = null;
            if( ! its[i].isFcClientItf() ) {
                Class<?> cl = loadFcInterceptClass(itsignature);
                intercept = instanciateFcInterceptClass(cl);
                ic.controllers.add(intercept);
                // For server interfaces, let the interceptor delegate to the content
                ((Interceptor)intercept).setFcItfDelegate(content);
            }
            
            // External interface
            Class<?> fcItfCl = loadFcItfClass(its[i]);
            proxy =
                instanciateFcItfClass(
                    fcItfCl, proxyForCompCtrl, itname, its[i], false,
                    intercept );

            // Internal interface
            if( ! itname.equals("attribute-controller") ) {
                InterfaceType intit = Utils.newSymetricInterfaceType(its[i]);
                fcItfCl = loadFcItfClass(intit);
                proxy =
                    instanciateFcItfClass(
                        fcItfCl, proxyForCompCtrl, itname, intit, true,
                        intercept );
                ic.internalInterfaces.put(itname,proxy);
            }
        }

        // Initialize controllers
        for( int i=0 ; i < ic.controllers.size() ; i++ ) {
            Object ctrl = ic.controllers.get(i);
            if( ctrl instanceof Controller ) {
                ((Controller)ctrl).initFcController(ic);
            }
        }
        
        // Return the fcinterface for the component controller
        return proxyForCompCtrl;
    }
    
    protected void contentChecks( ComponentType ct, Object content ) {
        
        // Indeed nothing

        /*
         * Generalization of the method for Juliac.
         */
//        if( content != null ) {
//            boolean checkForBC = false;
//            InterfaceType[] its = ct.getFcInterfaceTypes();
//            for (int i = 0; i < its.length; i++) {
//                if( its[i].isFcClientItf() ) {
//                    checkForBC = true;
//                }
//                else {
//                    String signature = its[i].getFcItfSignature();
//                    Class<?> cl = loadClassIE(signature);
//                    Class<?> contentcl = content.getClass();
//                    if( !(cl.isAssignableFrom(contentcl)) ) {
//                        String contentClassName = content.getClass().getName();                    
//                        String msg =
//                            contentClassName+" should implement "+signature;
//                        throw new InstantiationException(msg);                        
//                    }
//                }
//            }
//            if( checkForBC ) {
//                contentCheckForBC(content);
//            }
//        }        
    }

    protected void contentCheckForBC( Object content ) {
        
        // Indeed nothing

        /*
         * Generalization of the method for Juliac.
         */
//        if( content != null ) {
//            Class<?> contentcl = content.getClass();
//            if( !(BindingController.class.isAssignableFrom(contentcl)) ) {
//                String contentClassName = content.getClass().getName();                    
//                String msg =
//                    contentClassName+" should implement "+
//                    BindingController.class.getName();
//                throw new InstantiationException(msg);                        
//            }
//        }
    }
    
    protected void initializationContextForContent(
        InitializationContext ic, Object content, String ctrlDesc,
        Object contentDesc ) { 
        
        ic.content = contentDesc;

        /*
         * Generalization of the method for Juliac.
         */
//        ic.interfaces.put("/content",content);
//        ic.interfaces.put("/controllerDesc",ctrlDesc);
//
//        /*
//         * In the case of template components, Julia factory controller expects
//         * ic.content to reference contentDesc (which is an array describing
//         * the component associated to the template).
//         */
//        if( ctrlDesc.endsWith("Template") ) {
//            ic.content = contentDesc;
//        }
//        else {
//            ic.content = content;
//        }
    }
    
    /**
     * Load the class implementing the Fractal interface corresponding to the
     * given interface type. If the class can not be loaded from the classpath,
     * this method uses ASM to generate its bytecode and to dynamically load it.
     */
    protected Class<?> loadFcItfClass( InterfaceType it )
    throws InstantiationException {
        
        ClassGenerator iicg =
            new InterfaceImplementationClassGenerator(it,null,false,null);
        String name = iicg.getTargetClassName();
        
        try {
            Class<?> cl = loadClass(name);
            return cl;
        }
        catch (ClassNotFoundException e) {
            // TODO use ASM to generate the implementation of the Fractal interface
            throw new ChainedInstantiationException(e,null,"");
        }
    }
    
    /**
     * Instantiate the specified class with the specified constructor arguments.
     * The class must extend {@link Interface} and provide a <code>({@link
     * Component}, String,{@link Type},boolean,Object)</code> constructor.
     */
    protected Interface instanciateFcItfClass(
        Class<?> cl, Component owner, String name, Type type,
        boolean isInternal, Object impl )
    throws InstantiationException {
        
        try {
            Constructor<?> ctr =
                cl.getConstructor(
                    Component.class, String.class, Type.class, boolean.class,
                    Object.class );
            Interface itf = (Interface)
                ctr.newInstance(owner,name,type,isInternal,impl);
            return itf;
        }
        catch( Exception e ) {
            throw new ChainedInstantiationException(e,null,"");
        }
    }

    /**
     * Load the class implementing the Fractal interceptor corresponding to the
     * given Java interface fully-qualified name. If the class can not be loaded
     * directly from the classpath, this method uses ASM to generate its
     * bytecode and to dynamically load it.
     */
    protected Class<?> loadFcInterceptClass( String name )
    throws InstantiationException {
        
        /*
         * TODO critical problem: the suffix for scaComposite is InterceptorIntent.
         * We somehow need to revert to a solution where the interceptor source
         * code generator can be loaded to query it and to retrieve the name of
         * the generated class. This means that:
         * - either the interceptor classes must be splitted in two (along the
         *   ITBasedClassGenerator interceptor) to avoid depending on the full
         *   implementation of source code generator,
         * - or that the current module (tinfi-runtime-oo-dyn) must depend on
         *   tinfi-compiler and juliac-compiler to access these classes.
         *   Note 1: this wouldn't certainly be a big deal if no other module
         *           were necessary.
         *   Note 2: this removes the need to have a juliac-core module. 
         */
        name += "InterceptorLCSCACCIntent";
        
        try {
            Class<?> cl = loadClass(name);
            return cl;
        }
        catch (ClassNotFoundException e) {
            // TODO use ASM to generate the implementation of the Fractal interface
            throw new ChainedInstantiationException(e,null,"");
        }
    }
    
    /**
     * Instantiate the specified class with the specified constructor arguments.
     * The class must extend {@link Interface} and provide a <code>({@link
     * Component}, String,{@link Type},boolean,Object)</code> constructor.
     */
    protected Object instanciateFcInterceptClass( Class<?> cl )
    throws InstantiationException {
        
        try {
            Object intercept = cl.newInstance();
            return intercept;
        }
        catch( Exception e ) {
            throw new ChainedInstantiationException(e,null,"");
        }
    }
}
