/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.component;

import java.lang.reflect.Constructor;
import java.util.Stack;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.type.InterfaceType;
import org.osoa.sca.CallableReference;
import org.osoa.sca.ComponentContext;
import org.osoa.sca.RequestContext;
import org.osoa.sca.ServiceReference;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.control.content.SCAContentController;
import org.ow2.frascati.tinfi.control.property.SCAPropertyController;
import org.ow2.frascati.tinfi.osoa.ServiceReferenceImpl;

/**
 * Mixin layer for implementing the {@link ComponentContext} controller.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public abstract class ComponentContextMixin implements ComponentContext {

    // -------------------------------------------------------------------------
    // Private constructor
    // -------------------------------------------------------------------------

    private ComponentContextMixin () {}

    // -------------------------------------------------------------------------
    // Implementation of the ComponentContext interface
    // -------------------------------------------------------------------------

    /**
     * Returns the absolute URI of the component within the SCA Domain.
     *
     * @return the absolute URI of the component
     */
    public String getURI() {
        throw new UnsupportedOperationException();
    }

    /**
     * Cast a type-safe reference to a CallableReference.
     * Converts a type-safe reference to an equivalent CallableReference; if the target refers to a service
     * then a ServiceReference will be returned, if the target refers to a callback then a CallableReference
     * will be returned.
     *
     * @param target a reference proxy provided by the SCA runtime
     * @param <B> the Java type of the business interface for the reference
     * @param <R> the type of reference to be returned
     * @return a CallableReference equivalent for the proxy
     * @throws IllegalArgumentException if the supplied instance is not a reference supplied by the SCA runtime
     */
    public <B, R extends CallableReference<B>> R cast(B target)
    throws IllegalArgumentException {
        
        if( !(target instanceof ServiceReference) ) {
            String msg = "Not a reference: "+target;
            throw new IllegalArgumentException(msg);
        }
        
        ServiceReference<?> ci = (ServiceReference<?>) target;
        Class<?> businessInterface = ci.getBusinessInterface();
        Object service = ci.getService();
        
        // Create a ServiceReference
        String signature = businessInterface.getName();
        String clname = signature+"FcSR";

        try {
            Class<?> cl = Class.forName(clname);
            Constructor<?>[] ctrs = cl.getConstructors();
            Constructor<?> ctr = ctrs[0];
            Object sr = ctr.newInstance(businessInterface,service);
            return (R) sr;
        }
        catch (Exception e) {
            throw new IllegalArgumentException(e);
        }
    }

    /**
     * Returns a proxy for a reference defined by this component.
     *
     * @param businessInterface the interface that will be used to invoke the service
     * @param referenceName the name of the reference
     * @return an object that implements the business interface
     */
    public <B> B getService(Class<B> businessInterface, String referenceName)
    throws TinfiRuntimeException {
        
        try {
            Interface itf = (Interface) _this_weaveableC.getFcInterface(referenceName);
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( ! it.isFcClientItf() ) {
                String msg = "No such reference: "+referenceName;
                throw new TinfiRuntimeException(msg);
            }
            if( ! businessInterface.isAssignableFrom(itf.getClass()) ) {
                String msg = "Reference can not be casted to: "+businessInterface.getClass();
                throw new TinfiRuntimeException(msg);
            }
            return (B) itf;
        }
        catch (NoSuchInterfaceException e) {
            throw new TinfiRuntimeException(e);
        }
    }

    /**
     * Returns a ServiceReference for a reference defined by this component.
     *
     * @param businessInterface the interface that will be used to invoke the service
     * @param referenceName the name of the reference
     * @return a ServiceReference for the designated reference
     */
    public <B> ServiceReference<B> getServiceReference(Class<B> businessInterface, String referenceName) {
        B service = getService(businessInterface, referenceName);
        ServiceReference<B> sr = new ServiceReferenceImpl<B>(businessInterface,service);
        return sr;
    }

    /**
     * Returns the value of an SCA property defined by this component.
     *
     * @param type the Java type to be returned for the property
     * @param propertyName the name of the property whose value should be returned
     * @return the property value
     */
    public <B> B getProperty(Class<B> type, String propertyName)
    throws IllegalArgumentException {
        
        if( ! _this_weavablePC.containsKey(propertyName) ) {
            String msg = "No such property: "+propertyName;
            throw new IllegalArgumentException(msg);
        }
        
        Object value = _this_weavablePC.get(propertyName);
        
        try {
            return (B)value;
        }
        catch( ClassCastException cce ) {
            String msg = "Property "+propertyName+" is not of type "+type.getName();
            throw new IllegalArgumentException(msg);
        }
    }

    /**
     * Returns a ServiceReference that can be used to invoke this component over
     * the default service.
     *
     * @param businessInterface the interface that will be used to invoke the service
     * @return a ServiceReference that will invoke this component
     */
    public <B> ServiceReference<B> createSelfReference(Class<B> businessInterface) {
        B service = getServ(businessInterface);
        ServiceReference<B> sr = new ServiceReferenceImpl<B>(businessInterface,service);
        return sr;
    }

    /**
     * Returns a ServiceReference that can be used to invoke this component over
     * the designated service.
     *
     * @param businessInterface the interface that will be used to invoke the service
     * @param serviceName the name of the service to invoke
     * @return a ServiceReference that will invoke this component
     */
    public <B> ServiceReference<B> createSelfReference(Class<B> businessInterface, String serviceName) {
        B service = getServ(businessInterface, serviceName);
        ServiceReference<B> sr = new ServiceReferenceImpl<B>(businessInterface,service);
        return sr;
    }

    /**
     * Returns the context for the current SCA service request, or null if there
     * is no current request or if the context is unavailable.
     *
     * @return the SCA request context; may be null
     */
    public RequestContext getRequestContext() {
        Stack<RequestContext> stack = _this_weavableSCACC.getRequestContextStack();
        RequestContext top = stack.peek();
        return top;
    }

    
    // -------------------------------------------------------------------------
    // Implementation specific
    // -------------------------------------------------------------------------

    /**
     * Returns a proxy for the default service defined by this component.
     * The default service is defined as the first server interface found in the
     * list of interfaces associated with this component.
     *
     * @param businessInterface the interface that will be used to invoke the service
     * @param serviceName the name of the reference
     * @return an object that implements the business interface
     */
    private <B> B getServ(Class<B> businessInterface)
    throws TinfiRuntimeException {
        
        Object[] itfs = _this_weaveableC.getFcInterfaces();
        for (Object o : itfs) {
            Interface itf = (Interface) o;
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( ! it.isFcClientItf() ) {
                if( businessInterface.isAssignableFrom(itf.getClass()) ) {
                    return (B) itf;
                }
            }
        }
        
        String msg = "No such default service";
        throw new TinfiRuntimeException(msg);
    }

    /**
     * Returns a proxy for a service defined by this component.
     *
     * @param businessInterface the interface that will be used to invoke the service
     * @param serviceName the name of the reference
     * @return an object that implements the business interface
     */
    private <B> B getServ(Class<B> businessInterface, String serviceName)
    throws TinfiRuntimeException {
        
        try {
            Interface itf = (Interface) _this_weaveableC.getFcInterface(serviceName);
            InterfaceType it = (InterfaceType) itf.getFcItfType();
            if( it.isFcClientItf() ) {
                String msg = "No such service: "+serviceName;
                throw new TinfiRuntimeException(msg);
            }
            if( ! businessInterface.isAssignableFrom(itf.getClass()) ) {
                String msg = "Service can not be casted to: "+businessInterface.getClass();
                throw new TinfiRuntimeException(msg);
            }
            return (B) itf;
        }
        catch (NoSuchInterfaceException e) {
            throw new TinfiRuntimeException(e);
        }
    }


    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

    /**
     * The {@link Component} interface of the component to which this controller
     * object belongs.
     */
    public Component _this_weaveableC;

    /**
     * The {@link SCAPropertyController} interface of the component to which
     * this controller object belongs.
     */
    public SCAPropertyController _this_weavablePC;

    /**
     * The {@link SCAContentController} interface of the component to which
     * this controller object belongs.
     */
    public SCAContentController _this_weavableSCACC;
}
