/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2008-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Authors: Lionel Seinturier, Philippe Merle
 */

package org.ow2.frascati.tinfi.control.content;

import java.lang.annotation.Annotation;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.juliac.runtime.ClassLoaderFcItf;
import org.objectweb.fractal.juliac.runtime.ClassLoaderItf;
import org.objectweb.fractal.juliac.runtime.Juliac;
import org.osoa.sca.ComponentContext;
import org.osoa.sca.annotations.Callback;
import org.osoa.sca.annotations.ComponentName;
import org.osoa.sca.annotations.Context;
import org.osoa.sca.annotations.ConversationAttributes;
import org.osoa.sca.annotations.ConversationID;
import org.osoa.sca.annotations.Destroy;
import org.osoa.sca.annotations.EagerInit;
import org.osoa.sca.annotations.Init;
import org.osoa.sca.annotations.Property;
import org.osoa.sca.annotations.Reference;
import org.osoa.sca.annotations.Scope;
import org.ow2.frascati.tinfi.metadata.MetaData;

/**
 * This class stores SCA related metadata for the content class of a component.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Philippe Merle <Philippe.Merle@lifl.fr>
 */
public class ContentClassMetaData {

    // -----------------------------------------------------------------------
    // Factory
    // -----------------------------------------------------------------------
    
    /**
     * Retrieve the SCA related metadata for the specified component class name.
     * 
     * @param fcContentDesc  the component class name (so called content class)
     * @return               the corresponding metadata
     * @throws IllegalContentClassMetaData
     *      if an error occurs while retrieving metadata 
     */
    public static ContentClassMetaData get( String fcContentDesc )
    throws IllegalContentClassMetaData {
        
        ContentClassMetaData ccmd = ccmds.get(fcContentDesc);
        if( ccmd == null ) {
            Class<?> c = loadClass(fcContentDesc);
            ccmd = new ContentClassMetaData(c);
            ccmds.put(fcContentDesc,ccmd);
        }
        
        return ccmd;
    }
    
    /**
     * Retrieve the SCA related metadata for the specified component class name.
     * 
     * @param fcContentDesc  the component class name (so called content class)
     * @param cl             the class loader for loading the component class
     * @return               the corresponding metadata
     * @throws IllegalContentClassMetaData
     *      if an error occurs while retrieving metadata 
     */
    public static ContentClassMetaData get( String fcContentDesc, ClassLoader cl )
    throws IllegalContentClassMetaData {
        
        ContentClassMetaData ccmd = ccmds.get(fcContentDesc);
        if( ccmd == null ) {
            Class<?> c = loadClass(fcContentDesc,cl);
            ccmd = new ContentClassMetaData(c);
            ccmds.put(fcContentDesc,ccmd);
        }
        
        return ccmd;
    }
    
    private static Map<String,ContentClassMetaData> ccmds =
        new HashMap<String,ContentClassMetaData>();

    
    // -------------------------------------------------------------------------
    // Public data
    // -------------------------------------------------------------------------
    
    /** The component content class. */
    public Class<?> fcContentClass;
    
    /** The @{@link EagerInit} annotation, if any. */
    public EagerInit eagerinit;

    /** The scope policy, if any. */
    public Scope scope;
    
    /** The conversation attributes, if any. */
    public ConversationAttributes convattr;
    
    /** The maximun age of the conversation in milli seconds. */
    public long convMaxAge = 0;
    
    /** The maximun idle time of the conversation in milli seconds. */
    public long convMaxIdleTime = 0;
    
    /** The constructor annotated with @{@link Constructor}, if any. */
    public Constructor<?> constructorAnnotatedElement;
    
    /** The method annotated with @{@link Init}, if any. */
    public Method initMethod;
    
    /** The method annotated with @{@link Destroy}, if any. */
    public Method destroyMethod;
    
    /** The field or setter method annotated with @{@link Context}, if any. */
    public AnnotatedElement contextAnnotatedElement;
    
    /** The field or setter method annotated with @{@link ComponentName}, if any. */
    public AnnotatedElement componentNameAnnotatedElement;
    
    /** The field or setter method annotated with @{@link ConversationID}, if any. */
    public AnnotatedElement conversationIDAnnotatedElement;
    
    /** List of references. */
    public AnnotatedElementsManager refs;

    /** List of properties. */
    public AnnotatedElementsManager props;
    
    /** The unannotated setter methods. */
    public Map<String,Method> setters;
    
    /** The fields or methods annotated with @{@link Callback}. */
    public AccessibleObject[] callbacks;

    
    // -------------------------------------------------------------------------
    // Constructor to initialize data
    // -------------------------------------------------------------------------
    
    private MetaData<?> md;
    public MetaData<?> getMetaData() { return md; }
    
    /**
     * Retrieve the SCA related metadata for the specified component class name.
     * 
     * @throws IllegalContentClassMetaData
     *      if an error occurs while retrieving metadata 
     */
    private <T> ContentClassMetaData( Class<T> fcContentClass )
    throws IllegalContentClassMetaData {
        
        md = new MetaData<T>(fcContentClass);
        this.fcContentClass = fcContentClass;
        
        /*
         * Retrieve the scope policy of the content class.
         */
        scope = md.getAnnotation(Scope.class);
        
        /*
         * Retrieve the scope eager init policy of the content class, if any.
         */
        eagerinit = md.getAnnotation(EagerInit.class);
        if( eagerinit != null ) {
            if( scope==null || !scope.value().equals("COMPOSITE") ) {
                String msg =
                    "Class "+fcContentClass+
                    " annotated with @EagerInit should be composite-scoped";
                throw new IllegalContentClassMetaData(msg);
            }
        }

        /*
         * Retrieve the conversation attributes of the content class.
         */
        convattr = md.getAnnotation(ConversationAttributes.class);
        if( convattr != null ) {
            String ma = convattr.maxAge();
            convMaxAge = convDurationToMilli(ma);
            String mit = convattr.maxIdleTime();
            convMaxIdleTime = convDurationToMilli(mit);
        }
        
        /*
         * Retrieve the constructor annotated with @Constructor, if any.
         * Check that only one such constructor exists.
         */
        Constructor<?>[] constructors =
            md.getDeclaredAnnotatedConstructors(
                org.osoa.sca.annotations.Constructor.class );
        if( constructors.length > 1 ) {
            String msg = "Only one constructor should be annotated with @Constructor";
            throw new IllegalContentClassMetaData(msg);
        }
        constructorAnnotatedElement = (constructors.length==1) ? constructors[0] : null;

        /*
         * Retrieve the method annotated with @Init, if any.
         * Check that only one such method exists.
         */
        Method[] inits = md.getAllAnnotatedMethods(Init.class);
        initMethod = checkFcUniqueAnnotatedVoidMethod(inits,Init.class);

        /*
         * Retrieve the method annotated with @Destroy, if any.
         * Check that only one such method exists.
         */
        Method[] destroys = md.getAllAnnotatedMethods(Destroy.class);
        destroyMethod =
            checkFcUniqueAnnotatedVoidMethod(destroys,Destroy.class);

        /*
         * Retrieve the element annotated with @Context, if any.
         */
        AccessibleObject[] aos =
            md.getAllAnnotatedMethodsAndFields(Context.class);
        contextAnnotatedElement =
            checkFcUniqueAnnotatedElement(
                aos,Context.class,ComponentContext.class);
        
        /*
         * Retrieve the element annotated with @ComponentName, if any.
         */
        aos = md.getAllAnnotatedMethodsAndFields(ComponentName.class);
        componentNameAnnotatedElement =
            checkFcUniqueAnnotatedElement(aos,ComponentName.class,String.class);

        /*
         * Retrieve the element annotated with @ConversationID, if any.
         */
        aos = md.getAllAnnotatedMethodsAndFields(ConversationID.class);
        conversationIDAnnotatedElement =
            checkFcUniqueAnnotatedElement(
                aos,ConversationID.class,Object.class);

        /*
         * Retrieve the elements annotated with @Reference.
         */
        refs = new AnnotatedElementsManager(md,Reference.class);
        try {
            refs.addAllReferences();
        }
        catch (AnnotatedElementNotFoundException e) {
            throw new IllegalContentClassMetaData(e);
        }
        catch (DuplicateAnnotatedElementException e) {
            throw new IllegalContentClassMetaData(e);
        }
        
        /*
         * Retrieve the elements annotated with @Property.
         */
        props = new AnnotatedElementsManager(md,Property.class);
        try {
            props.addAllReferences();
        }
        catch (AnnotatedElementNotFoundException e) {
            throw new IllegalContentClassMetaData(e);
        }
        catch (DuplicateAnnotatedElementException e) {
            throw new IllegalContentClassMetaData(e);
        }
        
        /*
         * Retrieve the unannotated setter methods.
         */
        setters = md.getAllUnAnnotatedSetterMethods();
        
        /*
         * Retrieve the callbacks declared in the content class.
         */
        callbacks = md.getAllAnnotatedMethodsAndFields(Callback.class);
        for (AccessibleObject ao : callbacks) {
            if( ao instanceof Constructor ) {
                String msg =
                    "Illegal @Callback annotation on constructor in "+
                    fcContentClass;
                throw new IllegalContentClassMetaData(msg);
            }
            else if( ao instanceof Method ) {
                Method method = (Method) ao;
                Util.checkSetterMethodSignature(method,null,null);
            }
        }
    }

    
    // -------------------------------------------------------------------------
    // Implementation specific
    // -------------------------------------------------------------------------

    /**
     * Load the class whose name is specified with the specified classloader.
     */
    private static Class<?> loadClass( String name, ClassLoader cl )
    throws IllegalContentClassMetaData {
        try {
            Class<?> c = cl.loadClass(name);
            return c;
        }
        catch( ClassNotFoundException e ) {
            throw new IllegalContentClassMetaData(e);
        }
    }
    
    /**
     * Load the class whose name is specified with the classloader interface
     * provided by the Juliac bootstrap component.
     */
    private static Class<?> loadClass( String name )
    throws IllegalContentClassMetaData {

        ClassLoaderItf cl = null;
        try {
            Component boot = new Juliac().newFcInstance();
            cl = (ClassLoaderItf) boot.getFcInterface(ClassLoaderFcItf.NAME);
        }
        catch( NoSuchInterfaceException nsie ) {
            /*
             * Shouldn't occur since Juliac provides a classloader interface.
             */
            throw new RuntimeException(nsie);
        }
        
        try {
            Class<?> c = cl.loadClass(name);
            return c;
        }
        catch( ClassNotFoundException e ) {
            throw new IllegalContentClassMetaData(e);
        }
    }
    
    /**
     * Check that at most one element, method or field, is annotated with the
     * specified annotation and is of the specified type.
     * 
     * @return  the {@link ReferencedElement} instance associated to the field or method
     * @throws IllegalContentClassMetaData  if this is not the case
     */
    protected AnnotatedElement checkFcUniqueAnnotatedElement(
        AccessibleObject[] aos, Class<? extends Annotation> annot, Class<?> type )
    throws IllegalContentClassMetaData {
        
        checkFcUniqueness(aos,annot);
        
        AnnotatedElement annotatedElement = null;
        if( aos.length == 1 ) {
            if( aos[0] instanceof Method ) {
                Method meth = (Method) aos[0];
                Util.checkSetterMethodSignature(meth,null,type);
                annotatedElement = new AnnotatedElement(meth);
            }
            else {
                Field field = (Field) aos[0];
                if( ! type.isAssignableFrom(field.getType()) ) {
                    String msg =
                        annot.getName()+" annotated field "+field.getName()+
                        " should be of type "+type.getName();
                    throw new IllegalContentClassMetaData(msg);
                }
                annotatedElement = new AnnotatedElement(field);
            }
        }
        return annotatedElement;
    }

    /**
     * Check that at most one void returning, no argument, method is annotated
     * with the specified annotation in the given array of methods.
     * 
     * @return  the method or null if there is no such method
     * @throws IllegalContentClassMetaData  if this is not the case
     */
    protected Method checkFcUniqueAnnotatedVoidMethod(
        Method[] meths, Class<? extends Annotation> annot )
    throws IllegalContentClassMetaData {
        
        checkFcUniqueness(meths,annot);
        
        Method meth = (meths.length==1) ? meths[0] : null;
        if( meth != null ) {
            if( ! meth.getReturnType().equals(void.class) ||
                meth.getParameterTypes().length != 0 ) {
                String msg =
                    "The signature of the @"+annot+" annotated method ("+
                    meth.getName()+") should be ():void";
                throw new IllegalContentClassMetaData(msg);
            }
        }
        
        return meth;
    }
    
    /**
     * Check that only one element is annotated with the specified annotation in
     * the given array of elements.
     * 
     * @throws IllegalContentClassMetaData  if this is not the case
     */
    protected void checkFcUniqueness(
        AccessibleObject[] aos, Class<? extends Annotation> annot )
    throws IllegalContentClassMetaData {
        if( aos.length > 1 ) {
            String msg =
                "More than one element annotated with @"+annot+" in "+
                fcContentClass.getName();
            throw new IllegalContentClassMetaData(msg);
        }        
    }
    
    /**
     * Return the duration in milli seconds corresponding to the specified
     * string. The format of the string is defined in the SCA specifications.
     * 
     * @throws IllegalContentClassMetaData  if the string is badly formatted
     */
    private static long convDurationToMilli( String duration )
    throws IllegalContentClassMetaData {
        
        int space = duration.indexOf(' ');
        if( space == -1 ) {
            String msg = "No space in duration string: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        String numberStr = duration.substring(0,space);
        int number = Integer.parseInt(numberStr);
        if( number < 0 ) {
            String msg = "Negative duration: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        if( space+1 == duration.length() ) {
            String msg = "No time unit in: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        String unit = duration.substring(space+1);
        long l = number;
        if( unit.equals("seconds") )       l *= 1000;
        else if( unit.equals("minutes") )  l *= 60*1000;
        else if( unit.equals("hours") )    l *= 60*60*1000;
        else if( unit.equals("days") )     l *= 24*60*60*1000;
        else if( unit.equals("years") )    l *= 365*24*60*60*1000;
        else {
            String msg = "Unsupported time unit in: "+duration;
            throw new IllegalContentClassMetaData(msg);
        }
        
        return l;
    }
}
