/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2009 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.juliac;

import java.io.IOException;
import java.util.Map;

import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.ClassGenerator;
import org.objectweb.fractal.juliac.InitializerClassGenerator;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.JuliacConfig;
import org.objectweb.fractal.juliac.JuliacLoader;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.desc.NoSuchControllerDescriptorException;
import org.objectweb.fractal.juliac.proxy.InterceptorSourceCodeGenerator;
import org.objectweb.fractal.juliac.proxy.InterfaceImplementationClassGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.osoa.sca.annotations.Callback;

/**
 * This class generates the source code associated to Fractal components.
 * The membrane implementation and the initializer implementation are generated.
 * 
 * The content, the interceptors and the controllers are kept in separate
 * classes and the controllers are implemented with objects.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class FCOOCtrlSourceCodeGenerator
extends org.objectweb.fractal.juliac.opt.oo.FCOOCtrlSourceCodeGenerator {
    
    public FCOOCtrlSourceCodeGenerator() {
        super();
    }

    /** Configuration file for Tinfi. */
    public static final String DEFAULT_CONFIGURATION = "tinfi-bundled.cfg";

    /**
     * Initialize this generator.
     * 
     * The initialization process is kept separated from the constructor in
     * order to be able to customize the julia.config system property.
     * 
     * @param jc  the instance of Juliac using this source code generator
     */
    @Override
    public void init( Juliac jc ) throws IOException {

        JuliacLoader loader = jc.getJuliacConfig().getJuliacLoader();
        loader.loadConfigFile(DEFAULT_CONFIGURATION);
        
        super.init(jc);

        /*
         * Register component interceptor code generators.
         */
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.putInterceptorSourceCodeGenerator(
            "org.ow2.frascati.tinfi.asm.SCAContentInterceptorCodeGenerator",
            SCAContentInterceptorSourceCodeGenerator.class.getName() );
        jconf.putInterceptorSourceCodeGenerator(
            "org.ow2.frascati.tinfi.asm.SCAIntentInterceptorCodeGenerator",
            SCAIntentInterceptorSourceCodeGenerator.class.getName() );
    }

    /**
     * Perform consistency checks on the parameter transmitted to
     * {@link #generate(Type, Object, Object)}.
     * 
     * @param type            the component type
     * @param controllerDesc  the component controller descriptor
     * @param contentDesc     the component content descriptor
     * @throws IllegalArgumentException  in case of inconsistency
     */
    @Override
    protected void check( Type type, Object controllerDesc, Object contentDesc )
    throws IllegalArgumentException {
        
        super.check(type,controllerDesc,contentDesc);
        
        /*
         * Even though we extend the default OO optimization level source code
         * generator {@link
         * org.objectweb.fractal.juliac.opt.oo.FCOOCtrlSourceCodeGenerator} for
         * conveniently inheriting from already written code, the current
         * optimization level source code generator must only treat
         * scaPrimitive and scaComposite components. If we don't perform this
         * check, primitive, composite, etc. components are treated but the
         * generated code is not consistant.
         */
        String ctrlDesc = (String) controllerDesc;
        if( !ctrlDesc.equals("scaPrimitive") && !ctrlDesc.equals("scaComposite") ) {
            String msg =
                "Only scaPrimitive and scaComposite controller descriptors "+
                "are supported by optimization level "+getClass().getName();
            throw new IllegalArgumentException(msg);
        }
    }
    
    /**
     * Return the source code generators for the initializer classes associated
     * with this component source code generator.
     */
    @Override
    protected InitializerClassGenerator getInitializerSourceCodeGenerator() {
        return new InitializerOOCtrlClassGenerator();
    }

    /**
     * Return the source code generator for component interfaces.
     */
    @Override
    public ProxyClassGenerator getInterfaceSourceCodeGenerator( 
        InterfaceType it, MembraneDesc membraneDesc ) {

        String ctrlDesc = membraneDesc.getDescriptor();
        InterfaceSourceCodeGeneratorFactory iscgf =
            new InterfaceSourceCodeGeneratorFactory(it,ctrlDesc,membraneDesc,jc);
        ProxyClassGenerator pcg = iscgf.getInterfaceSourceCodeGenerator();
        return pcg;
    }
    
    /**
     * Generate the source code of component interfaces associated with the
     * specified interface type.
     */
    @Override
    protected void generateInterfaceImpl( InterfaceType it, String ctrlDesc )
    throws IOException {
        
        InterfaceSourceCodeGeneratorFactory iscgf =
            new InterfaceSourceCodeGeneratorFactory(it,ctrlDesc,null,jc);
        
        if( iscgf.isSCACompBusinessInterface() ) {
           
            /*
             * Generate SCA specific interface implementations.
             */

            // Server (input) interfaces
            ClassGenerator cg = iscgf.getInterfaceSourceCodeGenerator();
            jc.generateSourceCode(cg);
                
            // ServiceReference implementation
            cg = new ServiceReferenceClassGenerator(it,null,false,jc);
            jc.generateSourceCode(cg);
            
            // CallableReference and input interface implementations for callbacks
            String signature = it.getFcItfSignature();
            UnifiedClass cl = jc.create(signature);
            Callback annot = cl.getAnnotation(Callback.class);
            if( annot != null ) {
                Class<?> cbcl = annot.value();
                InterfaceType cbit =
                    new BasicInterfaceType(
                        "callback", cbcl.getName(), false, false, false );
                cg = new CallableReferenceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(cg);
                
                cg = new CallBackInterfaceClassGenerator(cbit,null,false,jc);
                jc.generateSourceCode(cg);
            }
        }
        else {
            
            /*
             * Interfaces of non-SCA components.
             */
            ClassGenerator cg = 
                new InterfaceImplementationClassGenerator(it,null,false,jc);
            jc.generateSourceCode(cg);
        }
    }

    /**
     * Set the interceptor source code generator for the specified membrane
     * descriptor.
     */
    @Override
    protected void setInterceptorSourceCodeGenerator(
        Map<?,?> context, MembraneDesc membraneDesc )
    throws
        NoSuchControllerDescriptorException, IllegalArgumentException,
        ClassNotFoundException {
        
        String desc = membraneDesc.getDescriptor();
        if( ! desc.startsWith("sca") ) {
            /*
             * For non SCA controller descriptors, use the regular
             * InterceptorClassGenerator class defined by the super method.
             */
            super.setInterceptorSourceCodeGenerator(context,membraneDesc);
            return;
        }
        
        // Get the interceptor source code generators from the Julia config file
        boolean mergeable = false;
        String ctrldesc = membraneDesc.getDescriptor();
        InterceptorSourceCodeGenerator[] scgs =
            getInterceptorSourceCodeGenerator(ctrldesc,context,mergeable);
        if( scgs == null ) {
            // No interceptor
            return;
        }
        
        // InterceptorClassGenerator specific for SCA components
        InterceptorClassGenerator icg = new InterceptorClassGenerator(scgs,jc);
        membraneDesc.setInterceptorClassGenerator(icg);
    }    
}
